<?php

namespace NoccyLabs\DevMode;

class Application
{

    protected $plugins = [];

    public function __construct()
    {
        $paths = [
            getcwd()."/.devmode",
            __DIR__."/../plugins",
            getenv("HOME")."/.local/share/devmode/plugins",
            getenv("HOME")."/.devmode/plugins",
            "/usr/share/devmode/plugins",
        ];

        $pm = new Plugin\PluginManager();
        $this->plugins = $pm->loadPlugins($paths);

    }

    public function run()
    {
        global $argv;
        list($opts,$args) = \Console_Getopt::getopt($argv, "hvV", [ "help", "verbose", "version" ]);

        foreach ($opts as list($opt,$optval)) {
            switch ($opt) {
                case 'h':
                case '--help':
                    $this->showUsage();
                    return EXIT_OK;
                case 'v':
                case '--verbose':
                    $verbosity = min($verbosity+1,3);
                    break;
                case 'V':
                case '--version':
                    $this->showVersion();
                    return EXIT_OK;
            }
        }

        if (count($args) == 0) {
            $this->showUsage();
            return EXIT_BAD_COMMAND;
        }

        $manifest = new Manifest();

        $action = array_shift($args);
        switch (strtolower($action)) {
            case 'up':
            case 'down':
            case 'status':
                return $this->action($manifest, strtolower($action))?:EXIT_OK;
            case 'setup':
            case 'teardown':
                return $this->provision($manifest, strtolower($action))?:EXIT_OK;
            case 'plugins':
                printf("There are %d plugins loaded:\n", count($this->plugins));
                foreach ($this->plugins as $plugin) {
                    $man = $plugin->getManifest();
                    printf(" + %s (v%s): %s\n", $man->getId(), $man->getVersion(), $man->getDescription());
                }
                return EXIT_OK;
            default:
                throw new \RuntimeException("Bad action: {$action}");
        }
    }

    private function showUsage()
    {
        printf("devmode [options] action [args...]\n");

        $options = [
            "-h,--help" => "Show this help",
            "-v[vv],--verbose" => "Verbose(r) output",
            "-V,--version" => "Print version and exit",
        ];
        printf("\nOptions:\n");
        foreach ($options as $option=>$info) {
            printf("  %-20s %s\n", $option, $info);
        }

        $actions = [
            "up"        => "Start all defined services",
            "down"      => "Stop all defined services",
            "status"    => "Show service status",
            "setup"     => "Provision requested resources",
            "teardown"  => "Clean up used resources",
            "plugins"   => "List loaded plugins",
        ];
        printf("\nActions:\n");
        foreach ($actions as $action=>$info) {
            printf("  %-15s %s\n", $action, $info);
        }

    }
    
    private function showVersion()
    {
        printf(APP_NAME." v".APP_VERSION.APP_VERSION_EXTRA."\n");
    }
    
    private function action(Manifest $manifest, $action)
    {
        $deps = $manifest->getDependencies();
        foreach ($deps as $dependency) {
            $dm = $this->getDependencyManager($dependency);
            if (!$dm) {
                fprintf(STDERR, "Could not find a dependency manager to handle [%s]\n", $dependency->getName());
                continue;
            }
            switch ($action) {
                case 'up':
                    $dm->setDependencyState($dependency, true);
                    break;
                case 'down':
                    $dm->setDependencyState($dependency, false);
                    break;
                case 'status':
                    $dm->showDependencyStatus($dependency);
                    break;
            }
        }
        return EXIT_OK;
    }

    private function provision(Manifest $manifest, $action)
    {
        $resources = $manifest->getResources();
        foreach ($resources as $resource) {
            $pm = $this->getProvisioner($resource);
            if (!$pm) {
                fprintf(STDERR, "Could not find a provisioner to handle [%s]\n", $resource->getName());
                continue;
            }
            switch ($action) {
                case 'setup':
                    $pm->provisionResource($resource, true);
                    break;
                case 'teardown':
                    $pm->provisionResource($resource, false);
                    break;
            }
        }
        return EXIT_OK;
    }

    public function getPluginsByInterface($interface)
    {
        $plugins = [];
        foreach ($this->plugins as $plugin) {
            if ($plugin instanceof $interface) {
                $plugins[] = $plugin;
            }
        }
        return $plugins;
    }
    
    private function getDependencyManager(Dependency $dependency)
    {
        $plugins = $this->getPluginsByInterface(Plugin\IDependencyManager::class);
        $manager = null; $managerWeight = null;
        foreach ($plugins as $plugin) {
            $w = $plugin->isManagedDependency($dependency);
            if ($w == false) {
                continue;
            }
            if ((float)$w > $managerWeight) {
                $manager = $plugin;
                $managerWeight = (float)$w;
            }
        }
        if (!$manager) {
            throw new \RuntimeException("Could not find a manager for {$dependency->getId()}");
        }
        return $manager;
    }

    private function getProvisioner(Resource $resource)
    {
        $plugins = $this->getPluginsByInterface(Plugin\IProvisioner::class);
        $manager = null; $managerWeight = null;
        foreach ($plugins as $plugin) {
            $w = $plugin->canProvisionResource($resource);
            if ($w == false) {
                continue;
            }
            if ((float)$w > $managerWeight) {
                $manager = $plugin;
                $managerWeight = (float)$w;
            }
        }
        if (!$manager) {
            throw new \RuntimeException("Could not find a manager for {$dependency->getId()}");
        }
        return $manager;
    }

}
