<?php

namespace NoccyLabs\DevMode\Plugin;

use NoccyLabs\DevMode\Dependency;

class PluginManager
{
    protected $plugins = [];

    public function loadPlugins(array $paths)
    {
        foreach ($paths as $path) {
            $this->pluginScan($path);
        }
        $loaded = $this->pluginLoad();

        return $loaded;
    }

    private function pluginScan($path)
    {
        $plugins = glob($path."/*");
        foreach ($plugins as $dir) {
            if (!file_exists($dir."/plugin.json")) {
                continue;
            }
            $manifest = new PluginManifest($dir."/plugin.json");
            $this->pluginRegister($manifest);
        }
        
    }

    private function pluginRegister(PluginManifest $manifest)
    {
        $id = $manifest->getId();

        if (array_key_exists($id, $this->plugins)) {
            if (!$manifest->isNewerVersion($this->plugins[$id])) {
                //printf("Skipping [%s] as its already registered in version (%s)\n", $id, $this->plugins[$id]->getVersion());
                return;
            }
        }
        $this->plugins[$id] = $manifest;

    }

    public function pluginLoad()
    {
        $plugins = array_values($this->plugins);
        $ordered = [];
        $loaded = [];
        while (count($plugins)) {
            $manifest = array_shift($plugins);
            $sorted = [];
            $added = false;
            foreach ($ordered as $compare) {
                if ($manifest->shouldLoadBefore($compare)) {
                    $sorted[] = $manifest;
                    $added = true;
                }
                $sorted[] = $compare;
                if ($manifest->shouldLoadAfter($compare)) {
                    $sorted[] = $manifest;
                    $added = true;
                }
            }
            if (!$added) {
                $sorted[] = $manifest;
            }
            $ordered = $sorted;
        }

        $ret = [];
        foreach ($ordered as $plugin) {
            $conflicts = false;
            foreach ($ret as $manifest) {
                if ($plugin->conflictsWith($manifest->getManifest())) {
                    $conflicts = true;
                }
            }
            if (!$conflicts) {
                $pluginClass = require_once $plugin->getPath()."/plugin.php";
                $pluginClass->setManifest($plugin);
                if ($pluginClass->onLoad() === false) {
                    continue;
                }
                $ret[$plugin->getId()] = $pluginClass;
                
            }
        }

        return $ret;
    }
}

class PluginManifest
{
    protected $id = null;

    protected $path = null;

    protected $version = null;

    protected $loadBefore = [];

    protected $loadAfter = [];

    protected $conflictsWith = [];

    protected $requires = [];

    protected $loaded = false;

    protected $instance = null;

    public function __construct($filename=null)
    {
        if ($filename) {
            $this->readManifest($filename);
        }
    }

    public function readManifest($filename)
    {
        $manifest = json_decode(file_get_contents($filename));
        $this->path = dirname($filename);

        $this->id = $manifest->plugin;
        $this->description = (string)@$manifest->description;
        $this->version = (string)@$manifest->version?:0;
        $this->conflictsWith = (array)@$manifest->conflicts;
        $this->loadBefore = (array)@$manifest->before;
        $this->loadAfter = (array)@$manifest->after;
        $this->requires = (array)@$manifest->requires;
    }

    public function getId()
    {
        return $this->id;
    }

    public function getPath()
    {
        return $this->path;
    }

    public function getVersion()
    {
        return $this->version;
    }

    public function getDescription()
    {
        return $this->description;
    }

    public function isNewerVersion(PluginManifest $plugin)
    {
        return (version_compare($this->version, $plugin->getVersion()) > 0);
    }

    public function shouldLoadBefore(PluginManifest $plugin)
    {
        foreach ($this->loadBefore as $rule) {
            if (fnmatch($rule, $plugin->getId()))
                return true;
        }
        return false;
    }

    public function shouldLoadAfter(PluginManifest $plugin)
    {
        foreach ($this->loadAfter as $rule) {
            if (fnmatch($rule, $plugin->getId()))
                return true;
        }
        return false;
    }

    public function conflictsWith(PluginManifest $plugin)
    {
        foreach ($this->conflictsWith as $rule) {
            if (fnmatch($rule, $plugin->getId()))
                return true;
        }
        return false;
    }

    public function getDependencies()
    {
        return $this->requires;
    }

}